package de.unihannover.orderingExperimentAnalysis;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.List;
import java.util.Random;
import java.util.stream.Collectors;

import de.unihannover.orderingExperimentUi.Diff;
import de.unihannover.orderingExperimentUi.Experiment;
import de.unihannover.orderingExperimentUi.Treatment;
import de.unihannover.orderingExperimentUi.Treatments;
import de.unihannover.orderingExperimentUi.Treatments.TreatmentCombination;

public class SimulateExperiment {

    private static Random random = new Random(1234);
    private static long idCounter;

    private static final class Participant {
        private final CsvRow baseData;
        private final double basicSkill;
        private final double timeFactor;

        public Participant(CsvRow baseData) {
            this.baseData = baseData;
            this.basicSkill = this.determineBasicSkill();
            this.timeFactor = this.determineTimeFactor();
        }

        public int getTestReviewDefects() {
            return this.getReviewDefects(this.basicSkill, 1, 0, 2);
        }

        public int getDefectsFor(Diff d1, Treatment t1, boolean difference) {
            double skill = this.basicSkill;
            if (difference) {
                switch (t1) {
                case OPTIMAL_FILES:
                    skill += 0.04;
                    break;
                case OPTIMAL_NO_FILES:
                    skill += 0.05;
                    break;
                case WORST_FILES:
                    skill -= 0.03;
                    break;
                case WORST_NO_FILES:
                    skill -= 0.07;
                    break;
                }
            }

            switch (d1) {
            case DIFF1:
                return this.getReviewDefects(skill, 5, 3, 1);
            case DIFF2:
                return this.getReviewDefects(skill, 2, 2, 4);
            default:
                throw new AssertionError();
            }
        }

        private double determineBasicSkill() {
            final int maxExp = Math.max(this.baseData.getI("experienceSoftwareDev"), this.baseData.getI("experienceJava"));
            final int maxPract = Math.max(this.baseData.getI("currentProgrammingPractice"), this.baseData.getI("currentReviewPractice"));
            final double baseScore =
                ((maxExp + maxPract) / 10.0) * 0.4
                + (this.baseData.getD("workingMemorySpan") / 80.0) * 0.4
                + (1.0 - (this.baseData.getI("workingHours") / 9.0)) * 0.15
                + (this.baseData.getI("screenHeight") / 1080.0) * 0.05;
            return baseScore * 0.5 + random.nextDouble() * 0.5;
        }

        private int getReviewDefects(double basicSkill, int easyCount, int mediumCount, int hardCount) {
            int cnt = 0;
            cnt += this.binomialSampling(basicSkill + 0.1, easyCount);
            cnt += this.binomialSampling(basicSkill, mediumCount);
            cnt += this.binomialSampling(basicSkill - 0.1, hardCount);
            return cnt;
        }

        private int binomialSampling(double skill, int count) {
            int found = 0;
            for (int i = 0; i < count; i++) {
                if (random.nextDouble() < skill) {
                    found++;
                }
            }
            return found;
        }

        public int getTestReviewTime() {
            return this.sampleNormal(3 * 60 * this.timeFactor);
        }

        public int getTimeFor(Diff d1, Treatment t1, boolean difference) {
            double meanTime;
            switch (d1) {
            case DIFF1:
                meanTime = 20 * 60;
                break;
            case DIFF2:
                meanTime = 25 * 60;
                break;
            default:
                throw new AssertionError();
            }

            if (difference) {
                switch (t1) {
                case OPTIMAL_FILES:
                    meanTime *= 0.92;
                    break;
                case OPTIMAL_NO_FILES:
                    meanTime *= 0.93;
                    break;
                case WORST_FILES:
                    meanTime *= 1.03;
                    break;
                case WORST_NO_FILES:
                    meanTime *= 1.05;
                    break;
                }
            }

            return this.sampleNormal(meanTime * this.timeFactor);
        }

        private double determineTimeFactor() {
            final int maxExp = Math.max(this.baseData.getI("experienceSoftwareDev"), this.baseData.getI("experienceJava"));
            final int maxPract = Math.max(this.baseData.getI("currentProgrammingPractice"), this.baseData.getI("currentReviewPractice"));
            final double baseScore =
                ((maxExp + maxPract) / 10.0) * 0.5
                + (this.baseData.getD("workingMemorySpan") / 80.0) * 0.3
                + (1.0 - (this.baseData.getI("workingHours") / 9.0)) * 0.05
                + (this.baseData.getI("screenHeight") / 1080.0) * 0.15;
            return 1.25 - baseScore * 0.5;
        }

        private int sampleNormal(double mean) {
            final double stddev = 0.2 * mean;
            final double sample = random.nextGaussian() * stddev + mean;
            return (int) Math.max(30, sample);
        }

        public String get(String col) {
            return this.baseData.get(col);
        }

    }

    public static void main(String[] args) throws Exception {
        final CsvTable exps = CsvReader.readRStyle(new File("40experiments.csv"));
        final CsvTable pars = CsvReader.readRStyle(new File("40participants.csv"));

        final Writer resultExps = new FileWriter(new File("resultE.csv"));
        resultExps.write("id;withDifference\n");

        final Writer resultPars = new FileWriter(new File("resultP.csv"));
        resultPars.write("experimentId");
        wf(resultPars, "experienceSoftwareDev");
        wf(resultPars, "experienceJava");
        wf(resultPars, "currentProgrammingPractice");
        wf(resultPars, "currentReviewPractice");
        wf(resultPars, "workingHours");
        wf(resultPars, "hasExperience");
        wf(resultPars, "screenHeight");
        wf(resultPars, "workingMemorySpan");

        wf(resultPars, "testReviewTime");
        wf(resultPars, "testReviewDefects");

        wf(resultPars, "r1Diff");
        wf(resultPars, "r1Trtmt");
        wf(resultPars, "r1Time");
        wf(resultPars, "r1Defects");

        wf(resultPars, "r2Diff");
        wf(resultPars, "r2Trtmt");
        wf(resultPars, "r2Time");
        wf(resultPars, "r2Defects");

        resultPars.write('\n');

        int sum = 0;
        final List<Participant> parList = pars.getRows().stream().map(Participant::new).collect(Collectors.toList());
        for (final CsvRow exp : exps.getRows()) {
            final int participantCount = exp.getI("participantCount");
            simulateExperiments(parList.subList(sum, sum + participantCount), resultExps, resultPars);
            sum += participantCount;
        }
        resultExps.close();
        resultPars.close();
    }

    private static void simulateExperiments(List<Participant> participants, Writer resultExps, Writer resultPars) throws IOException {
        simulateExperiment(participants, resultExps, resultPars, false);
        simulateExperiment(participants, resultExps, resultPars, true);
    }

    private static void simulateExperiment(
                    List<Participant> participants, Writer resultExps, Writer resultPars, boolean difference) throws IOException {
        final long curId = idCounter++;
        resultExps.write(curId + ";" + (difference ? "T" : "F") + "\n");

        final Treatments t = new Treatments(new Random(curId));
        for (final Participant p : participants) {
            final Experiment exp = new Experiment(0);
            resultPars.write(Long.toString(curId));
            wf(resultPars, p.get("experienceSoftwareDev"));
            wf(resultPars, p.get("experienceJava"));
            wf(resultPars, p.get("currentProgrammingPractice"));
            wf(resultPars, p.get("currentReviewPractice"));
            wf(resultPars, p.get("workingHours"));
            wf(resultPars, p.get("hasExperience"));
            wf(resultPars, p.get("screenHeight"));
            wf(resultPars, p.get("workingMemorySpan"));

            final int testReviewScore = p.getTestReviewDefects();
            wf(resultPars, p.getTestReviewTime());
            wf(resultPars, testReviewScore);

            exp.setTestReviewScore(testReviewScore);
            final TreatmentCombination tc = exp.getAssignedTreatment(t);

            wf(resultPars, tc.getD1());
            wf(resultPars, tc.getT1());
            wf(resultPars, p.getTimeFor(tc.getD1(), tc.getT1(), difference));
            wf(resultPars, p.getDefectsFor(tc.getD1(), tc.getT1(), difference));

            wf(resultPars, tc.getD2());
            wf(resultPars, tc.getT2());
            wf(resultPars, p.getTimeFor(tc.getD2(), tc.getT2(), difference));
            wf(resultPars, p.getDefectsFor(tc.getD2(), tc.getT2(), difference));

            resultPars.write('\n');
        }
    }

    private static void wf(Writer w, Object o) throws IOException {
        w.write(";" + o);
    }

}
